@@ -1,10 +1,11 @@
-import re, sys, json, math, time, os, uuid, muid
+import re, sys, json, math, time, os, uuid, itertools
 import pymorton
 from itertools import zip_longest
 import numpy as np
 from redis.client import list_or_args
 from typing import List, Union, Any, Optional
-from microprediction.conventions import MicroConventions
+from microconventions import MicroConventions
+from rediz.samplers import exponential_bootstrap
 
 KeyList   = List[Optional[str]]
 NameList  = List[Optional[str]]
@@ -14,8 +15,8 @@ DelayList = List[Optional[int]]
 
 SEP = "::"
 
-REDIZ_CONVENTIONS_ARGS = ('history_len', 'lagged_len', 'delays', 'max_ttl', 'error_ttl', 'transactions_ttl','error_limit', 'num_predictions','windows','obscurity','delay_grace','instant_recall')
-MICRO_CONVENTIONS_ARGS = ('min_len','min_balance')
+REDIZ_CONVENTIONS_ARGS = ('history_len', 'delays','lagged_len', 'max_ttl', 'error_ttl', 'transactions_ttl','error_limit', 'num_predictions','windows','obscurity','delay_grace','instant_recall')
+MICRO_CONVENTIONS_ARGS = ('num_predictions','min_len','min_balance','delays')
 
 class RedizConventions(MicroConventions):
 
@@ -27,22 +28,22 @@ class RedizConventions(MicroConventions):
 
         if windows is None:
             windows = [1e-4, 1e-3,  1e-2]
-        self.SEP = SEP
+        self.SEP = SEP  # TODO: Redundent now
         self.COPY_SEP = self.SEP + "copy" + self.SEP
         self.PREDICTION_SEP = self.SEP + "prediction" + self.SEP
 
         # User facing conventions: transparent use of prefixing
+        # TODO: These are moved to MicroConventions ... can delete this soon
+
         self.DELAYED = "delayed" + self.SEP
         self.CDF = 'cdf'+self.SEP
         self.LINKS = "links" + self.SEP
         self.BACKLINKS = "backlinks" + self.SEP
         self.MESSAGES = "messages" + self.SEP
         self.HISTORY = "history" + self.SEP
-        self.HISTORY_LEN = int( history_len or 1000)
         self.LAGGED = "lagged"+self.SEP
         self.LAGGED_VALUES = "lagged_values" + self.SEP
         self.LAGGED_TIMES = "lagged_times" + self.SEP
-        self.LAGGED_LEN = int( lagged_len or 10000)
         self.SUBSCRIBERS = "subscribers" + self.SEP
         self.SUBSCRIPTIONS = "subscriptions" + self.SEP
         self.TRANSACTIONS = "transactions" + self.SEP
@@ -51,21 +52,29 @@ class RedizConventions(MicroConventions):
         self.BALANCE = "balance" + self.SEP
         self.PERFORMANCE = "performance" + self.SEP
         self.LEADERBOARD = "leaderboard" + self.SEP
+        self.CUSTOM_LEADERBOARD = 'custom_leaderboard' + self.SEP
         self.BUDGETS = "budget" + self.SEP
         self.VOLUMES = "volumes" + self.SEP
         self.SUMMARY = "summary" + self.SEP
 
+        # Transparent but parametrized
+        self.HISTORY_LEN = int(history_len or 1000)
+        self.LAGGED_LEN = int(lagged_len or 10000)
+
         # Logging
-        self.CONFIRMS = "confirms" + self.SEP
-        self.WARNINGS = "warnings" + self.SEP
-        self.ERRORS   = "errors" + self.SEP
+        self.CONFIRMS = "confirms" + self.SEP  # TODO: Moved to MicroConventions
+        self.WARNINGS = "warnings" + self.SEP  # TODO: Moved to MicroConventions
+        self.ERRORS   = "errors" + self.SEP    # TODO: Moved to MicroConventions
+
+
         self.WARNINGS_TTL = int(60 * 60)  # TODO: allow configuation
         self.WARNINGS_LIMIT = 1000
         self.CONFIRMS_TTL = int(error_ttl or 60 * 60)  # Number of seconds that set execution error logs are persisted
-        self.ERROR_TTL = int(error_ttl or 60 * 60)  # Number of seconds that set execution error logs are persisted
+        self.ERROR_TTL = int(error_ttl or 60 * 60)     # Number of seconds that set execution error logs are persisted
         self.CONFIRMS_TTL = int(error_ttl or 60 * 60)  # Number of seconds that set execution error logs are persisted
-        self.ERROR_LIMIT = int(error_limit or 1000)  # Number of error messages to keep per write key
-        self.CONFIRMS_LIMIT = int(error_limit or 1000)  # Number of error messages to keep per write key
+        self.ERROR_LIMIT = int(error_limit or 1000)    # Number of error messages to keep per write key
+        self.CONFIRMS_LIMIT = int(error_limit or 1000) # Number of error messages to keep per write key
+        self.TRANSACTIONS_LIMIT = 1000
 
         # User transparent temporal and other config
         self.MIN_LEN = int(self.min_len)                     # FIXME: Get rid of MIN_LEN and use min_len instead
@@ -89,8 +98,9 @@ class RedizConventions(MicroConventions):
         self._SAMPLES = self._obscurity + self.SAMPLES  # Prefix to delayed predictions by horizon. Contain write_keys !
         self._PROMISED = "promised" + self.SEP  # Prefixes temporary values referenced by the promise queue
         self._DONATIONS = "donations" + self.SEP  # Write key donations
-        self._DONATION_PASSWORD = muid.shash(self._obscurity)  # Write key donation password
+        self._DONATION_PASSWORD = MicroConventions.shash(self._obscurity)  # Write key donation password
         self._DONORS = "donors"
+        self._DISCOUNT = 0.9            # Transfers
 
         # Other implementation config
         self._DELAY_GRACE = int(delay_grace or 5)  # Seconds beyond the schedule time when promise data expires
@@ -99,7 +109,7 @@ class RedizConventions(MicroConventions):
         self._INSTANT_RECALL = instant_recall or False
         self._MAX_TTL = int( max_ttl or 60 * 60 ) # Maximum TTL, useful for testing
         self._TRANSACTIONS_TTL = int( transactions_ttl or 24 * (60 * 60) )  # How long to keep transactions stream for inactive write_keys
-
+        self._LEADERBOARD_TTL  = int( 24 * (60 * 60)*60 )  # How long to keep transactions stream for inactive write_keys
 
     @staticmethod
     def assert_not_in_reserved_namespace(names, *args):
@@ -108,7 +118,7 @@ class RedizConventions(MicroConventions):
             raise Exception("Operation attempted with a name that uses " + RedizConventions.sep())
 
 
-    @staticmethod
+    @staticmethod  # TODO: Remove this after pushing MicroConventions
     def is_vector_value(value):
         if isinstance(value, (list, tuple)):
             return all((RedizConventions.is_scalar_value(v) for v in value))
@@ -119,7 +129,7 @@ class RedizConventions(MicroConventions):
             except:
                 return False
 
-    @staticmethod
+    @staticmethod # TODO: Remove this after pushing MicroConventions
     def is_dict_value(value):
         try:
             d = dict(value)
@@ -131,7 +141,7 @@ class RedizConventions(MicroConventions):
             except:
                 return False
 
-    @staticmethod
+    @staticmethod # TODO: Remove this after pushing MicroConventions
     def to_record(value):
         if RedizConventions.is_scalar_value(value):
             fields = {"0": value}
@@ -164,6 +174,7 @@ class RedizConventions(MicroConventions):
         write_keys = write_keys or [ write_key for _ in names ]
         return names, values, write_keys, budgets
 
+    # TODO: Move these ----------
     @staticmethod
     def delay_as_int(delay):
         """ By convention, None means no delay """
@@ -172,7 +183,9 @@ class RedizConventions(MicroConventions):
     def percentile_name(self, name, delay):
         return self.zcurve_name(names=[name],delay=delay)
 
-    def donation_name(self,len):
+
+
+    def donation_name(self,len):  # TODO: Move this elsewhere
         return self._DONATIONS + str(len)
 
     def donors_name(self):
@@ -191,7 +204,7 @@ class RedizConventions(MicroConventions):
         return self.CONFIRMS + write_key + '.json'
 
     def errors_name(self, write_key):
-        return self.MESSAGES + write_key + '.json'
+        return self.ERRORS + write_key + '.json'
 
     def warnings_name(self, write_key):
         return self.WARNINGS + write_key + '.json'
@@ -210,37 +223,79 @@ class RedizConventions(MicroConventions):
     def balance_name(self, write_key):
         return self.BALANCE + write_key + '.json'
 
-    def performance_key(self, name, delay):
-        name_stem = os.path.splitext(name)[0]
-        return name_stem + self.SEP + str(delay)
+    def horizon_name(self, name, delay):
+        """ Convention is used for performance and other hashes """  # Move to MicroConventions
+        return str(delay)+self.SEP + name
 
-    def leaderboard_name(self, name=None, delay=None):
+    def split_horizon_name(self, key):
+        spl = key.split(self.SEP)
+        name = spl[1]
+        delay = int(spl[0])
+        return name, delay
+
+    def split_horizon_names(self, keys):
+        names_delays = [self.split_horizon_name(key) for key in keys]
+        names = [n for n, _ in names_delays]
+        delays = [d for _, d in names_delays]
+        return names, delays
+
+    def custom_leaderboard_name(self, sponsor, name=None, dt=None):
+        """ Names for leaderboards with a given sponsor
+        :param sponsor:  str
+        :param name:     str
+        :param dt:       datetime
+        :return:
+        """
+        def lb_cat(name=None):
+            if name is not None:
+                if 'z1~' in name:
+                    return 'zscores_univariate'
+                elif 'z2~' in name:
+                    return 'zcurves_bivariate'
+                elif 'z3~' in name:
+                    return 'zcurves_trivariate'
+                else:
+                    return 'regular'
+            else:
+                return 'all_streams'
+
+        def lb_month(dt=None):
+            return dt.isoformat()[:7] if dt is not None else 'all_time'
+
+        sponsor = sponsor or 'overall'
+        return self.SEP.join([self.CUSTOM_LEADERBOARD[:-2], sponsor.replace(' ','_'), lb_cat(name), lb_month(dt)]) + '.json'
+
+
+    def leaderboard_name(self, name=None, delay=None): # TODO: discard after MicroConventions upgrade
+        """ Name for leaderboards by stream name and horizon """
         if name is None and delay is None:
             return self.LEADERBOARD[:-2]+'.json'
         elif name is None:
             return self.LEADERBOARD+str(delay)+'.json'
+        elif delay is None:
+            return self.LEADERBOARD+str(name)
         else:
-            return self.LEADERBOARD+name if delay is None else self.LEADERBOARD+str(delay)+self.SEP+name
+            return self.LEADERBOARD+self.horizon_name(name=name,delay=delay)
 
-    def history_name(self, name):
+    def history_name(self, name):   # TODO: discard after MicroConventions upgrade
         return self.HISTORY + name
 
-    def lagged_values_name(self, name):
+    def lagged_values_name(self, name):   # TODO: discard after MicroConventions upgrade
         return self.LAGGED_VALUES + name
 
-    def lagged_times_name(self, name):
+    def lagged_times_name(self, name):   # TODO: discard after MicroConventions upgrade
         return self.LAGGED_TIMES + name
 
-    def links_name(self, name, delay):
+    def links_name(self, name, delay):   # TODO: discard after MicroConventions upgrade
         return self.LINKS + str(delay) + self.SEP + name
 
-    def backlinks_name(self, name):
+    def backlinks_name(self, name):   # TODO: discard after MicroConventions upgrade
         return self.BACKLINKS + name
 
-    def subscribers_name(self, name):
+    def subscribers_name(self, name):   # TODO: discard after MicroConventions upgrade
         return self.SUBSCRIBERS + name
 
-    def subscriptions_name(self, name):
+    def subscriptions_name(self, name):   # TODO: discard after MicroConventions upgrade
         return self.SUBSCRIPTIONS + name
 
 
@@ -285,8 +340,9 @@ class RedizConventions(MicroConventions):
         return {"delayed": self.delayed_name,
                 "links": self.links_name}
 
-    def cdf_name(self,name,delay=None):
-        return self.CDF + name if delay==None else self.CDF+str(delay)+self.SEP+name
+    def cdf_name(self, name, delay=None):
+        return self.CDF + name if delay == None else self.CDF + str(delay) + self.SEP + name
+
 
 
 
@@ -342,7 +398,7 @@ class RedizConventions(MicroConventions):
     def _make_scenario_obscure(self, ticket):
         """ Change write_key to a hash of write_key """
         parts = ticket.split(self.SEP)
-        return parts[0] + self.SEP + muid.shash(parts[1])
+        return parts[0] + self.SEP + self.shash(parts[1])
 
     def _scenario_percentile(self, scenario):
         """ Extract scenario percentile from scenario string """
@@ -433,21 +489,21 @@ class RedizConventions(MicroConventions):
     #           Statistics
     # --------------------------------------------------------------------------
 
-    def normcdf(self, x):
+    def normcdf(self, x):  # TODO: Discard after MicroConv upgrade
         g = self._normcdf_function()
         return g(x)
 
-    def norminv(self,p):
+    def norminv(self,p): # TODO: Discard after MicroConv upgrade
         f = self._norminv_function()
         return f(p)
 
-    @staticmethod
+    @staticmethod  # TODO: discard as this is in MicroConventions now
     def to_zscores(prctls):
         norminv = RedizConventions._norminv_function()
         return [ norminv(p) for p in prctls ]
 
 
-    @staticmethod
+    @staticmethod # TODO: discard as this is in MicroConventions now
     def _norminv_function():
         try:
             from statistics import NormalDist
@@ -456,7 +512,7 @@ class RedizConventions(MicroConventions):
             from scipy.stats import norm
             return norm.ppf
 
-    @staticmethod
+    @staticmethod   # TODO: discard as this is in MicroConventions now
     def _normcdf_function():
         try:
             from statistics import NormalDist
@@ -465,8 +521,8 @@ class RedizConventions(MicroConventions):
             from scipy.stats import norm
             return norm.cdf
 
-    @staticmethod
-    def _zmean_percentile(ps):
+    @staticmethod  # TODO: discard as this is in MicroConventions now
+    def zmean_percentile(ps):
         """ Given a vector of percentiles, returns normal percentile of the mean zscore """
         if len(ps):
             norminv = RedizConventions._norminv_function()
@@ -479,36 +535,36 @@ class RedizConventions(MicroConventions):
             return 0.5
 
     # --------------------------------------------------------------------------
-    #           Z-order curves
+    #           Z-order curves  TODO: Moved all this to MicroConventions
     # --------------------------------------------------------------------------
 
-    def zcurve_names(self, names):
+    def zcurve_names(self, names):   # TODO: Discard after MicroConv upgrade
         import itertools
         znames=list()
         for delay in self.DELAYS:
            for dim in [1,2,3]:
-                name_combinations = itertools.combinations(sorted(names),dim)
+                name_combinations = list(itertools.combinations(sorted(names),dim))
                 zname = self.zcurve_name( names=name_combinations,delay=delay )
                 znames.append(zname)
         return znames
 
-    def zcurve_name(self, names, delay, obscure=False):
+    def zcurve_name(self, names, delay): # TODO: Discard after MicroConv upgrade
         """ Naming convention for derived quantities, called zcurves """
         basenames = sorted( [n.split('.')[-2] for n in names] )
         prefix    = "z" + str(len(names))
         clearbase = "~".join( [prefix] + basenames + [str(delay)] )
-        return clearbase+'.json' if not obscure else RedizConventions.hash(clearbase)+'.json'
+        return clearbase+'.json'
 
-    @staticmethod
+    @staticmethod # TODO: Discard after MicroConv upgrade
     def morton_scale(dim):
         return 2**10
 
     @staticmethod
-    def morton_large(dim):
+    def morton_large(dim): # TODO: Discard after MicroConv upgrade
         SCALE = RedizConventions.morton_scale(dim=dim)
         return pymorton.interleave( *[ SCALE-1 for _ in range(dim) ] )
 
-    def to_zcurve(self, prctls: List[float] ):
+    def to_zcurve(self, prctls: List[float] ): # TODO: Discard after MicroConv upgrade
         """ A mapping from R^n -> R based on the Morton z-curve """
         dim = len(prctls)
         if dim==1:
@@ -523,7 +579,7 @@ class RedizConventions(MicroConventions):
             zpercentile =  m1/m2
             return self.norminv(zpercentile)
 
-    def from_zcurve(self, zvalue, dim):
+    def from_zcurve(self, zvalue, dim): # TODO: Discard after MicroConv upgrade
         zpercentile = self.normcdf(zvalue)
         SCALE = self.morton_scale(dim)
         zmorton     = int( self.morton_large(dim)*zpercentile+0.5 )
@@ -541,7 +597,6 @@ class RedizConventions(MicroConventions):
     # --------------------------------------------------------------------------
 
     def empirical_predictions(self, lagged_values ):
-        from rediz.samplers import exponential_bootstrap
         predictions = exponential_bootstrap(lagged=lagged_values, decay=0.005, num=self.NUM_PREDICTIONS)
         return sorted(predictions)
 
